'''
Run scrapy spiders from a script.

Blocks and runs all requests in parallel.  Accumulated items from all
spiders are returned as a list.
'''

import collections
import inspect
from billiard import Process  # fork of multiprocessing that works with celery
from billiard.queues import Queue

from pydispatch import dispatcher
from scrapy import signals
from scrapy.crawler import CrawlerProcess
from scrapy.settings import Settings
from scrapy.spiders import Spider


class ScrapyScriptException(Exception):
    pass


class Job(object):
    '''A job is a single request to call a specific spider. *args and **kwargs
    will be passed to the spider constructor.
    '''

    def __init__(self, spider, *args, **kwargs):
        '''Parms:
          spider (spidercls): the spider to be run for this job.
        '''
        self.spider = spider
        self.args = args
        self.kwargs = kwargs


class Processor(Process):
    ''' Start a twisted reactor and run the provided scrapy spiders.
    Blocks until all have finished.
    '''

    def __init__(self, settings=None, item_scraped=True):
        '''
        Parms:
          settings (scrapy.settings.Settings) - settings to apply.  Defaults
        to Scrapy default settings.
        '''
        kwargs = {'ctx': __import__('billiard.synchronize')}

        self.results = Queue(**kwargs)
        self.counts = Queue(**kwargs)
        self.items = {}
        self.items_count = {}
        self.settings = settings or Settings()
        self.item_scraped = item_scraped
        dispatcher.connect(self._item_passed, signals.item_passed)

    def _item_passed(self, item, response, spider):
        if spider.name not in self.items.keys():
            self.items[spider.name] = []
            self.items_count[spider.name] = 0
        if self.item_scraped is True:
            self.items[spider.name].append(dict(item))

        self.items_count[spider.name] += 1

    def _crawl(self, requests):
        '''
        Parameters:
            requests (Request) - One or more Jobs. All will
                                 be loaded into a single invocation of the reactor.
        '''
        self.crawler = CrawlerProcess(self.settings)

        # crawl can be called multiple times to queue several requests
        for req in requests:
            self.crawler.crawl(req.spider, *req.args, **req.kwargs)

        self.crawler.start()
        self.crawler.stop()
        self.results.put(self.items)
        self.counts.put(self.items_count)

    def run(self, jobs):
        '''Start the Scrapy engine, and execute all jobs.  Return consolidated results
        in a single list.

        Parms:
          jobs ([Job]) - one or more Job objects to be processed.

        Returns:
          List of objects yielded by the spiders after all jobs have run.
        '''
        if not isinstance(jobs, collections.Iterable):
            jobs = [jobs]
        self.validate(jobs)

        p = Process(target=self._crawl, args=[jobs])
        p.start()
        p.join()
        p.terminate()

    def data(self):
        return self.results.get()

    def count(self):
        return self.counts.get()

    def validate(self, jobs):
        if not all([isinstance(x, Job) for x in jobs]):
            raise ScrapyScriptException('scrapy-script requires Job objects.')



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='scrapy-script',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.0.0',

    description='Run a Scrapy spider programmatically from a script or a Celery task - no project required.',
    long_description='scrapyscript allows you to invoke one or more spiders from a script, have them all run in parallel, and get the results back as a single list.  No scrapy project, no boilerplate, no hassle.',

    # The project's main homepage.
    url='https://github.com/aox-lei/scrapyscript',

    # Author details
    author='Aox lei',
    author_email='2387813033@qq.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Framework :: Scrapy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License'
    ],

    # What does your project relate to?
    keywords='scrapy',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=[],

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    py_modules=['scrapy_script'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['Scrapy>=1.4.0',
                      'billiard>=3.5.0.2',
    ],

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
    },
)
